import os

from relsad.network.components import MainController
from relsad.network.systems import PowerSystem
from relsad.results.storage import save_history
from relsad.Time import Time, TimeUnit
from relsad.visualization.plotting import plot_history, plot_history_last_state


def update_sequence_history(
    power_system: PowerSystem,
    prev_time: Time,
    curr_time: Time,
    save_flag: bool,
):
    """
    Updates the sequence history variables in the power system

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    prev_time : Time
        The previous time
    curr_time : Time
        The current time
    save_flag : bool
        Indicates if saving is on/off

    Returns
    ----------
    None

    """
    for network in power_system.child_network_list:
        for bus in network.buses:
            network.p_energy_shed += bus.p_energy_shed_stack
            network.q_energy_shed += bus.q_energy_shed_stack
        power_system.p_energy_shed += network.p_energy_shed
        power_system.q_energy_shed += network.q_energy_shed
        network.acc_p_energy_shed += network.p_energy_shed
        network.acc_q_energy_shed += network.q_energy_shed
    power_system.acc_p_energy_shed += power_system.p_energy_shed
    power_system.acc_q_energy_shed += power_system.q_energy_shed
    if save_flag:
        power_system_state_dict = {
            "p_energy_shed": power_system.p_energy_shed,
            "q_energy_shed": power_system.p_energy_shed,
            "acc_p_energy_shed": power_system.acc_p_energy_shed,
            "acc_q_energy_shed": power_system.acc_q_energy_shed,
            "p_load": power_system.get_system_load()[0],
            "q_load": power_system.get_system_load()[1],
        }
        for state_var, value in power_system_state_dict.items():
            power_system.history[state_var][curr_time] = value
        for network in power_system.child_network_list:
            network_state_dict = {
                "p_energy_shed": network.p_energy_shed,
                "q_energy_shed": network.q_energy_shed,
                "acc_p_energy_shed": network.acc_p_energy_shed,
                "acc_q_energy_shed": network.acc_q_energy_shed,
                "p_load": network.get_system_load()[0],
                "q_load": network.get_system_load()[1],
            }
            for state_var, value in network_state_dict.items():
                network.history[state_var][curr_time] = value
    power_system.p_energy_shed = 0
    power_system.q_energy_shed = 0
    for network in power_system.child_network_list:
        network.p_energy_shed = 0
        network.q_energy_shed = 0
    for comp in power_system.comp_list:
        comp.update_history(prev_time, curr_time, save_flag)
    power_system.controller.update_history(prev_time, curr_time, save_flag)


def plot_line_history(lines, save_dir: str):
    """
    Plots the history of the line in the power system

    Parameters
    ----------
    lines : list
        List of Line elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "p_from",
        "q_from",
        "p_to",
        "q_to",
        "remaining_outage_time",
        "failed",
        "line_loading",
    ]
    for state_var in whole_state_list:
        plot_history(lines, state_var, save_dir)


def save_line_history(lines, save_dir: str):
    """
    Saves the history of the line in the power system

    Parameters
    ----------
    lines : list
        List of Line elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "p_from",
        "q_from",
        "p_to",
        "q_to",
        "remaining_outage_time",
        "failed",
        "line_loading",
    ]
    for state_var in whole_state_list:
        save_history(lines, state_var, save_dir)


def plot_circuitbreaker_history(circuitbreakers, save_dir: str):
    """
    Plots the history of the circuitbreakers in the power system

    Parameters
    ----------
    circuitbreakers : list
        List of CircuitBreaker elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "is_open",
    ]
    for state_var in whole_state_list:
        plot_history(circuitbreakers, state_var, save_dir)


def save_circuitbreaker_history(circuitbreakers, save_dir: str):
    """
    Saves the history of the circuitbreakers in the power system

    Parameters
    ----------
    circuitbreakers : list
        List of CircuitBreaker elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "is_open",
    ]
    for state_var in whole_state_list:
        save_history(circuitbreakers, state_var, save_dir)


def plot_disconnector_history(disconnectors, save_dir: str):
    """
    Plots the history of the disconnectors in the power system

    Parameters
    ----------
    disconnectors: list
        List of Disconnector elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "is_open",
    ]
    for state_var in whole_state_list:
        plot_history(disconnectors, state_var, save_dir)


def save_disconnector_history(disconnectors, save_dir: str):
    """
    Saves the history of the disconnectors in the power system

    Parameters
    ----------
    disconnectors : list
        List of Disconnector elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "is_open",
    ]
    for state_var in whole_state_list:
        save_history(disconnectors, state_var, save_dir)


def plot_intelligent_switch_history(intelligent_switches, save_dir: str):
    """
    Plots the history of the intelligent switches in the power system

    Parameters
    ----------
    intelligent_switches : list
        List of IntelligentSwitch elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "remaining_repair_time",
        "state",
    ]
    for state_var in whole_state_list:
        plot_history(intelligent_switches, state_var, save_dir)


def save_intelligent_switch_history(intelligent_switches, save_dir: str):
    """
    Saves the history of the intelligent switches in the power system

    Parameters
    ----------
    intelligent_switches : list
        List of IntelligentSwitch elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "remaining_repair_time",
        "state",
    ]
    for state_var in whole_state_list:
        save_history(intelligent_switches, state_var, save_dir)


def plot_sensor_history(sensors, save_dir: str):
    """
    Plots the history of the sensors in the power system

    Parameters
    ----------
    sensors : list
        List of Sensor elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "remaining_repair_time",
        "state",
    ]
    for state_var in whole_state_list:
        plot_history(sensors, state_var, save_dir)


def save_sensor_history(sensors, save_dir: str):
    """
    Saves the history of the sensors in the power system

    Parameters
    ----------
    sensors : list
        List of Sensor elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "remaining_repair_time",
        "state",
    ]
    for state_var in whole_state_list:
        save_history(sensors, state_var, save_dir)


def plot_bus_history(buses, save_dir: str):
    """
    Plots the history of the buses in the power system

    Parameters
    ----------
    buses : list
        List of Bus elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "pload",
        "qload",
        "pprod",
        "qprod",
        "remaining_outage_time",
        "trafo_failed",
        "p_energy_shed_stack",
        "q_energy_shed_stack",
        "voang",
        "vomag",
    ]
    for state_var in whole_state_list:
        plot_history(buses, state_var, save_dir)
    last_state_list = [
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "avg_fail_rate",
        "avg_outage_time",
        "acc_outage_time",
        "interruption_fraction",
    ]
    for state_var in last_state_list:
        plot_history_last_state(buses, state_var, save_dir)


def save_bus_history(buses, save_dir: str):
    """
    Saves the history of the buses in the power system

    Parameters
    ----------
    buses : list
        List of Bus elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "pload",
        "qload",
        "pprod",
        "qprod",
        "remaining_outage_time",
        "trafo_failed",
        "p_energy_shed_stack",
        "q_energy_shed_stack",
        "voang",
        "vomag",
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "avg_fail_rate",
        "avg_outage_time",
        "acc_outage_time",
        "interruption_fraction",
    ]
    for state_var in whole_state_list:
        save_history(buses, state_var, save_dir)


def plot_ev_park_history(ev_parks, save_dir: str):
    """
    Plots the history of the ev_park in the power system

    Parameters
    ----------
    ev_parks : list
        List of EVPark elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "SOC",
        "ev_index",
        "demand",
        "charge",
        "num_cars",
        "park_interruption_fraction",
        "acc_num_interruptions",
        "acc_exp_interruptions",
        "acc_exp_car_interruptions",
        "acc_interruption_duration",
        "acc_available_num_cars",
    ]
    for state_var in whole_state_list:
        plot_history(ev_parks, state_var, save_dir)


def save_ev_park_history(ev_parks, save_dir: str):
    """
    Saves the history of the ev_park in the power system

    Parameters
    ----------
    ev_parks : list
        List of EVPark elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "SOC",
        "ev_index",
        "demand",
        "charge",
        "num_cars",
        "park_interruption_fraction",
        "acc_num_interruptions",
        "acc_exp_interruptions",
        "acc_exp_car_interruptions",
        "acc_interruption_duration",
        "acc_available_num_cars",
    ]
    for state_var in whole_state_list:
        save_history(ev_parks, state_var, save_dir)


def plot_battery_history(batteries, save_dir: str):
    """
    Plots the history of the battery in the power system

    Parameters
    ----------
    batteries : list
        List of Battery elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "SOC",
        "SOC_min",
        "remaining_survival_time",
    ]
    for state_var in whole_state_list:
        plot_history(batteries, state_var, save_dir)


def save_battery_history(batteries, save_dir: str):
    """
    Saves the history of the battery in the power system

    Parameters
    ----------
    batteries : list
        List of Battery elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "SOC",
        "SOC_min",
        "remaining_survival_time",
    ]
    for state_var in whole_state_list:
        save_history(batteries, state_var, save_dir)


def plot_power_system_history(power_system: PowerSystem, save_dir: str):
    """
    Plots the history of the power system

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "p_energy_shed",
        "q_energy_shed",
        "p_load",
        "q_load",
    ]
    power_system_save_dir = os.path.join(save_dir, power_system.name)
    for state_var in whole_state_list:
        plot_history([power_system], state_var, power_system_save_dir)
    for network in power_system.child_network_list:
        network_save_dir = os.path.join(save_dir, network.name)
        for state_var in whole_state_list:
            plot_history([network], state_var, network_save_dir)
    last_state_list = [
        "acc_p_energy_shed",
        "acc_q_energy_shed",
    ]
    power_system_save_dir = os.path.join(save_dir, power_system.name)
    for state_var in last_state_list:
        plot_history_last_state(
            [power_system], state_var, power_system_save_dir
        )
    for network in power_system.child_network_list:
        network_save_dir = os.path.join(save_dir, network.name)
        for state_var in last_state_list:
            plot_history_last_state([network], state_var, network_save_dir)


def save_power_system_history(power_system: PowerSystem, save_dir: str):
    """
    Saves the history of the power system

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "p_energy_shed",
        "q_energy_shed",
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "p_load",
        "q_load",
    ]
    power_system_save_dir = os.path.join(save_dir, power_system.name)
    for state_var in whole_state_list:
        save_history([power_system], state_var, power_system_save_dir)
    for network in power_system.child_network_list:
        network_save_dir = os.path.join(save_dir, network.name)
        for state_var in whole_state_list:
            save_history([network], state_var, network_save_dir)


def plot_network_controller_history(controllers, save_dir: str):
    """
    Plots the history of the controllers in the power system

    Parameters
    ----------
    controllers : list
        List of
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "sectioning_time",
    ]
    for state_var in whole_state_list:
        plot_history(controllers, state_var, save_dir)


def save_network_controller_history(controllers, save_dir: str):
    """
    Saves the history of the controllers in the power system

    Parameters
    ----------
    controllers : list
        A power system element
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "sectioning_time",
    ]
    for state_var in whole_state_list:
        save_history(controllers, state_var, save_dir)


def plot_system_controller_history(controllers, save_dir: str):
    """
    Plots the history of the controllers in the power system

    Parameters
    ----------
    controllers : list
        A power system element
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "sectioning_time",
        "remaining_repair_time",
        "state",
    ]
    for state_var in whole_state_list:
        plot_history(controllers, state_var, save_dir)


def save_system_controller_history(controllers, save_dir: str):
    """
    Saves the history of the controllers in the power system

    Parameters
    ----------
    controllers : list
        A power system element
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "sectioning_time",
        "remaining_repair_time",
        "state",
    ]
    for state_var in whole_state_list:
        save_history(controllers, state_var, save_dir)


def plot_ict_line_history(lines, save_dir: str):
    """
    Plots the history of the ICT line

    Parameters
    ----------
    lines : list
        List of ICTLine elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "remaining_outage_time",
        "failed",
    ]
    for state_var in whole_state_list:
        plot_history(lines, state_var, save_dir)


def save_ict_line_history(lines, save_dir: str):
    """
    Saves the history of the ICT line

    Parameters
    ----------
    lines : list
        List of ICTLine elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "remaining_outage_time",
        "failed",
    ]
    for state_var in whole_state_list:
        save_history(lines, state_var, save_dir)


def plot_ict_node_history(nodes, save_dir: str):
    """
    Plots the history of the ICT node

    Parameters
    ----------
    nodes : list
        List of ICTNode elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "remaining_outage_time",
        "failed",
        "avg_fail_rate",
        "avg_outage_time",
        "acc_outage_time",
        "acc_interruptions",
    ]
    for state_var in whole_state_list:
        plot_history(nodes, state_var, save_dir)


def save_ict_node_history(nodes, save_dir: str):
    """
    Saves the history of the ICT node

    Parameters
    ----------
    nodes : list
        List of ICTNode elements
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    whole_state_list = [
        "remaining_outage_time",
        "failed",
        "avg_fail_rate",
        "avg_outage_time",
        "acc_outage_time",
        "acc_interruptions",
    ]
    for state_var in whole_state_list:
        save_history(nodes, state_var, save_dir)


def initialize_sequence_history(power_system: PowerSystem):
    """
    Initializes the lists used for sequence history variables

    Parameters
    ----------
    power_system : PowerSystem
        A power system element

    Returns
    ----------
    None

    """
    network_state_list = [
        "p_energy_shed",
        "q_energy_shed",
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "p_load",
        "q_load",
    ]
    for state_var in network_state_list:
        power_system.history[state_var] = {}
    for network in power_system.child_network_list:
        for state_var in network_state_list:
            network.history[state_var] = {}
    power_system.controller.initialize_history()


def save_sequence_history(
    power_system: PowerSystem,
    save_dir: str,
):
    """
    Saves the history from an sequence

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    save_dir : str
        The saving path

    Returns
    ----------
    None

    """
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    save_power_system_history(
        power_system,
        os.path.join(save_dir),
    )
    save_bus_history(power_system.buses, os.path.join(save_dir, "bus"))
    if len(power_system.ev_parks) > 0:
        save_ev_park_history(
            power_system.ev_parks, os.path.join(save_dir, "ev_parks")
        )
    if len(power_system.batteries) > 0:
        save_battery_history(
            power_system.batteries, os.path.join(save_dir, "battery")
        )
    save_line_history(
        power_system.lines,
        os.path.join(save_dir, "line"),
    )
    if len(power_system.circuitbreakers) > 0:
        save_circuitbreaker_history(
            power_system.circuitbreakers,
            os.path.join(save_dir, "circuitbreaker"),
        )
    if len(power_system.disconnectors) > 0:
        save_disconnector_history(
            power_system.disconnectors,
            os.path.join(save_dir, "disconnector"),
        )

    if len(power_system.intelligent_switches) > 0:
        save_intelligent_switch_history(
            power_system.intelligent_switches,
            os.path.join(save_dir, "intelligent_switch"),
        )

    if len(power_system.sensors) > 0:
        save_sensor_history(
            power_system.sensors,
            os.path.join(save_dir, "sensor"),
        )

    if len(power_system.controller.distribution_controllers) > 0:
        save_network_controller_history(
            power_system.controller.distribution_controllers,
            os.path.join(save_dir, "distribution_controllers"),
        )

    if len(power_system.controller.microgrid_controllers) > 0:
        save_network_controller_history(
            power_system.controller.microgrid_controllers,
            os.path.join(save_dir, "microgrid_controllers"),
        )

    if isinstance(power_system.controller, MainController):
        save_system_controller_history(
            [power_system.controller],
            os.path.join(save_dir, "main_controller"),
        )

    if len(power_system.ict_lines) > 0:
        save_ict_line_history(
            power_system.ict_lines, os.path.join(save_dir, "ict_line")
        )

    if len(power_system.ict_nodes) > 0:
        save_ict_node_history(
            power_system.ict_nodes, os.path.join(save_dir, "ict_node")
        )



import copy
import os

from relsad.network.systems import PowerSystem
from relsad.reliability.indices import (
    ASAI,
    ASUI,
    CAIDI,
    ENS,
    SAIDI,
    SAIFI,
    EV_Duration,
    EV_Index,
    EV_Interruption,
)
from relsad.results.storage import save_monte_carlo_history_from_dict
from relsad.Time import Time
from relsad.visualization.plotting import plot_monte_carlo_history


def plot_network_monte_carlo_history(power_system: PowerSystem, save_dir: str):
    """
    Plots the history of the energy.shedding in the power system

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    save_dir : str
        The saving path

    Returns
    ----------
    None

    """
    network_state_list = [
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "SAIFI",
        "SAIDI",
        "CAIDI",
        "ASAI",
        "ASUI",
        "ENS",
        "EV_Index",
        "EV_Interruption",
        "EV_Duration",
    ]
    power_system_save_dir = os.path.join(save_dir, power_system.name)
    for state_var in network_state_list:
        plot_monte_carlo_history(
            [power_system], state_var, power_system_save_dir
        )
    for network in power_system.child_network_list:
        network_save_dir = os.path.join(save_dir, network.name)
        for state_var in network_state_list:
            plot_monte_carlo_history([network], state_var, network_save_dir)
    bus_state_list = [
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "avg_outage_time",
        "acc_outage_time",
        "interruption_fraction",
        "acc_interruptions",
    ]
    for bus in power_system.buses:
        bus_save_dir = os.path.join(save_dir, bus.name)
        for state_var in bus_state_list:
            plot_monte_carlo_history([bus], state_var, bus_save_dir)


def save_network_monte_carlo_history(
    power_system: PowerSystem, save_dir: str, save_dict: dict
):
    """
    Saves the history of the energy.shedding in the power system

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    save_dir : str
        The saving path
    save_dict : dict
        Dictionary with simulation results

    Returns
    ----------
    None

    """
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    network_state_list = [
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "SAIFI",
        "SAIDI",
        "CAIDI",
        "ASAI",
        "ASUI",
        "ENS",
        "EV_Index",
        "EV_Interruption",
        "EV_Duration",
    ]
    power_system_save_dir = os.path.join(save_dir, power_system.name)
    for state_var in network_state_list:
        save_monte_carlo_history_from_dict(
            save_dict, [power_system], state_var, power_system_save_dir
        )
    for network in power_system.child_network_list:
        network_save_dir = os.path.join(save_dir, network.name)
        for state_var in network_state_list:
            save_monte_carlo_history_from_dict(
                save_dict, [network], state_var, network_save_dir
            )
    bus_state_list = [
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "avg_outage_time",
        "acc_outage_time",
        "interruption_fraction",
        "acc_interruptions",
    ]
    ev_park_state_list = [
        "acc_num_interruptions",
        "acc_exp_interruptions",
        "acc_exp_car_interruptions",
        "acc_interruption_duration",
        "acc_available_num_cars",
        "num_cars",
    ]
    for bus in power_system.buses:
        bus_save_dir = os.path.join(save_dir, bus.name)
        for state_var in bus_state_list:
            save_monte_carlo_history_from_dict(
                save_dict, [bus], state_var, bus_save_dir
            )
        if bus.ev_park is not None:
            ev_park_save_dir = os.path.join(bus_save_dir, bus.ev_park.name)
            for state_var in ev_park_state_list:
                save_monte_carlo_history_from_dict(
                    save_dict, [bus.ev_park], state_var, ev_park_save_dir
                )


def initialize_monte_carlo_history(power_system: PowerSystem):
    """
    Initializes the lists used for history variables from the Monte Carlo simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    network_state_list = [
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "SAIFI",
        "SAIDI",
        "CAIDI",
        "ASAI",
        "ASUI",
        "ENS",
        "EV_Index",
        "EV_Interruption",
        "EV_Duration",
    ]
    save_dict = {}
    save_dict[power_system.name] = {}
    for state_var in network_state_list:
        save_dict[power_system.name][state_var] = {}
    for network in power_system.child_network_list:
        save_dict[network.name] = {}
        for state_var in network_state_list:
            save_dict[network.name][state_var] = {}
    bus_state_list = [
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "avg_outage_time",
        "acc_outage_time",
        "interruption_fraction",
        "acc_interruptions",
    ]
    ev_park_state_list = [
        "acc_num_interruptions",
        "acc_exp_interruptions",
        "acc_exp_car_interruptions",
        "acc_interruption_duration",
        "acc_available_num_cars",
        "num_cars",
    ]
    for bus in power_system.buses:
        save_dict[bus.name] = {}
        for state_var in bus_state_list:
            save_dict[bus.name][state_var] = {}
        if bus.ev_park is not None:
            save_dict[bus.ev_park.name] = {}
            for state_var in ev_park_state_list:
                save_dict[bus.ev_park.name][state_var] = {}
    return save_dict


def update_monte_carlo_power_system_history(
    power_system: PowerSystem,
    it: int,
    current_time: Time,
    save_dict: dict,
):
    """
    Updates the history dictionary from the Monte Carlo simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    it : int
        The iteration number
    current_time : Time
        Current time
    save_dict : dict
        Dictionary with simulation results

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    network_state_dict = {
        "acc_p_energy_shed": power_system.acc_p_energy_shed,
        "acc_q_energy_shed": power_system.acc_q_energy_shed,
        "SAIFI": SAIFI(power_system),
        "SAIDI": SAIDI(power_system),
        "CAIDI": CAIDI(power_system),
        "ASAI": ASAI(power_system, current_time),
        "ASUI": ASUI(power_system, current_time),
        "ENS": ENS(power_system),
        "EV_Index": EV_Index(power_system),
        "EV_Interruption": EV_Interruption(power_system),
        "EV_Duration": EV_Duration(power_system),
    }
    for state_var, value in network_state_dict.items():
        save_dict[power_system.name][state_var][it] = value
    save_dict = update_monte_carlo_child_network_history(
        power_system, it, current_time, save_dict
    )
    save_dict = update_monte_carlo_comp_history(power_system, it, save_dict)
    return save_dict


def update_monte_carlo_child_network_history(
    power_system: PowerSystem,
    it: int,
    current_time: Time,
    save_dict: dict,
):
    """
    Updates the history dictionary for the child networks in the Monte Carlo simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    it : int
        The iteration number
    current_time : Time
        Current time
    save_dict : dict
        Dictionary with simulation results

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    for network in power_system.child_network_list:
        network_state_dict = {
            "acc_p_energy_shed": network.acc_p_energy_shed,
            "acc_q_energy_shed": network.acc_q_energy_shed,
            "SAIFI": SAIFI(network),
            "SAIDI": SAIDI(network),
            "CAIDI": CAIDI(network),
            "ASAI": ASAI(network, current_time),
            "ASUI": ASUI(network, current_time),
            "ENS": ENS(network),
            "EV_Index": EV_Index(network),
            "EV_Interruption": EV_Interruption(network),
            "EV_Duration": EV_Duration(network),
        }
        for state_var, value in network_state_dict.items():
            save_dict[network.name][state_var][it] = value
    return save_dict


def update_monte_carlo_comp_history(
    power_system: PowerSystem,
    it: int,
    save_dict: dict,
):
    """
    Updates the component values for the system from the Monte Carlo simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    it : int
        The iteration number
    save_dict : dict
        Dictionary with simulation results

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    for bus in power_system.buses:
        bus_state_dict = {
            "acc_p_energy_shed": bus.acc_p_energy_shed,
            "acc_q_energy_shed": bus.acc_q_energy_shed,
            "avg_outage_time": bus.avg_outage_time,
            "acc_outage_time": bus.acc_outage_time,
            "interruption_fraction": bus.interruption_fraction,
            "acc_interruptions": bus.acc_interruptions,
        }
        for state_var, value in bus_state_dict.items():
            save_dict[bus.name][state_var][it] = value
        if bus.ev_park is not None:
            ev_park_state_dict = {
                "acc_num_interruptions": bus.ev_park.acc_num_interruptions,
                "acc_exp_interruptions": bus.ev_park.acc_exp_interruptions,
                "acc_exp_car_interruptions": bus.ev_park.acc_exp_car_interruptions,
                "acc_interruption_duration": bus.ev_park.acc_interruption_duration.get_hours(),
                "acc_available_num_cars": bus.ev_park.acc_available_num_cars,
                "num_cars": bus.ev_park.num_cars,
            }
            for state_var, value in ev_park_state_dict.items():
                save_dict[bus.ev_park.name][state_var][it] = value
    return save_dict


def merge_monte_carlo_history(
    power_system: PowerSystem,
    iteration_dicts: list,
):
    """
    Merges the Monte Carlo history from all the iterations in the simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    iteration_dicts : list
        List containing information about all the iterations

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    save_dict = copy.deepcopy(iteration_dicts[0])
    network_state_list = [
        "acc_p_energy_shed",
        "acc_q_energy_shed",
        "SAIFI",
        "SAIDI",
        "CAIDI",
        "ASAI",
        "ASUI",
        "ENS",
        "EV_Index",
        "EV_Interruption",
        "EV_Duration",
    ]
    for it_dict in iteration_dicts:
        it = list(it_dict[power_system.name][network_state_list[0]].keys())[0]
        for state_var in network_state_list:
            save_dict[power_system.name][state_var][it] = it_dict[
                power_system.name
            ][state_var][it]
        save_dict = merge_monte_carlo_child_network_history(
            power_system, it_dict, it, save_dict
        )
        save_dict = merge_monte_carlo_comp_history(
            power_system, it_dict, it, save_dict
        )
    return save_dict


def merge_monte_carlo_child_network_history(
    power_system: PowerSystem,
    it_dict: dict,
    it: int,
    save_dict: dict,
):
    """
    Initializes the lists used for history variables from the Monte Carlo simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    it_dict : dict
        Dictionary with iteration results
    it : int
        Iteration number
    save_dict : dict
        Dictionary with simulation results

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    for network in power_system.child_network_list:
        network_state_list = [
            "acc_p_energy_shed",
            "acc_q_energy_shed",
            "SAIFI",
            "SAIDI",
            "CAIDI",
            "ASAI",
            "ASUI",
            "ENS",
            "EV_Index",
            "EV_Interruption",
            "EV_Duration",
        ]
        for state_var in network_state_list:
            save_dict[network.name][state_var][it] = it_dict[network.name][
                state_var
            ][it]
    return save_dict


def merge_monte_carlo_comp_history(
    power_system: PowerSystem,
    it_dict: dict,
    it: int,
    save_dict: dict,
):
    """
    Initializes the lists used for history variables from the Monte Carlo simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    it_dict : dict
        Dictionary with iteration results
    it : int
        The iteration number
    save_dict : dict
        Dictionary with simulation results

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    for bus in power_system.buses:
        bus_state_list = [
            "acc_p_energy_shed",
            "acc_q_energy_shed",
            "avg_outage_time",
            "acc_outage_time",
            "interruption_fraction",
            "acc_interruptions",
        ]
        for state_var in bus_state_list:
            save_dict[bus.name][state_var][it] = it_dict[bus.name][state_var][
                it
            ]
        if bus.ev_park is not None:
            ev_park_state_list = [
                "acc_num_interruptions",
                "acc_exp_interruptions",
                "acc_exp_car_interruptions",
                "acc_interruption_duration",
                "acc_available_num_cars",
                "num_cars",
            ]
            for state_var in ev_park_state_list:
                save_dict[bus.ev_park.name][state_var][it] = it_dict[
                    bus.ev_park.name
                ][state_var][it]
    return save_dict



import matplotlib.lines as mlines
import numpy as np

from relsad.network.containers import SectionState
from relsad.Time import Time, TimeUnit
from relsad.utils import unique

from .Component import Component
from .Line import Line
from .Switch import Switch


class CircuitBreaker(Component, Switch):

    """
    Common base class for circuit breakers

    ...

        Attributes
        ----------
        name : string
            Name of the circuit breaker
        line : Line
            The line the circuit breaker is connected to
        is_open : bool
            Indicates if the circuit breaker is open or closed
        fail_rate : float
            The failure rate of the circuit breaker [no. of fails per year]
        outage_time : Time
            The outage time of the circuit breaker
        coordinate : list
            Coordinate of the circuit breaker
        initial_state : bool
            The initial state of the circuit breaker
        failed : bool
            True if the circuit breaker is in a failed state, False if not
        history : dict
            Dictonary attribute that stores the historic variables

        Methods
        ----------
        close()
            Closes the circuit breaker and the disconnectors connected to the host line
        open()
            Opens the circuit breaker and the disconnectors connected to the host line
        update_fail_status(dt)
            Updates the failure status of the circuit breaker
        initialize_history()
            Initializes the history variables
        update_history(prev_time, curr_time, save_flag)
            Updates the history variables
        get_history(attribute)
            Returns the history variables of an attribute
        add_random_instance(random_gen)
            Adds global random seed
        print_status()
            Prints the status
        reset_status(save_flag)
            Resets and sets the status of the system parameters


    """

    ## Visual attributes
    color = "black"
    edgecolor = "black"
    marker = "s"
    size = 3**2
    handle = mlines.Line2D(
        [],
        [],
        marker=marker,
        markeredgewidth=3,
        markersize=size,
        linestyle="None",
        color=color,
        markeredgecolor=edgecolor,
    )

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        line: Line,
        is_open: bool = False,
        fail_rate: float = 0.0,
        outage_time: Time = Time(1, TimeUnit.HOUR),
    ):

        # Verify input
        if line is None:
            raise Exception("CircuitBreaker must be connected to a Line")
        if line.parent_network is not None:
            raise Exception(
                "CircuitBreaker must be created before the line is connected to a network"
            )
        if fail_rate < 0:
            raise Exception("The failure rate must be positive")

        self.name = name

        dx = line.tbus.coordinate[0] - line.fbus.coordinate[0]
        dy = line.tbus.coordinate[1] - line.fbus.coordinate[1]
        self.coordinate = [
            line.fbus.coordinate[0] + dx / 3,
            line.fbus.coordinate[1] + dy / 3,
        ]
        self.initial_state = is_open
        self.is_open = is_open
        self.failed = False
        self.fail_rate = fail_rate
        self.outage_time = outage_time
        self.line = line

        # Add this circuitbreaker to parent line
        self.line.circuitbreaker = self

        ## History
        self.history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"CircuitBreaker(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(
                other, CircuitBreaker
            )
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def close(self):
        """
        Closes the circuit breaker and the disconnectors connected to the circuit breaker

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_open = False
        self.color = "black"
        for discon in self.line.disconnectors:
            if (
                discon.is_open
                and self.line.section.state == SectionState.CONNECTED
            ):
                discon.close()
        if not self.line.connected:
            self.line.connect()

    def open(self):
        """
        Opens the circuit breaker and the disconnectors connected to the circuit breaker

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_open = True
        self.color = "white"
        for discon in self.line.disconnectors:
            if not discon.is_open:
                discon.open()
        if self.line.connected:
            self.line.disconnect()

    def update_fail_status(self, dt: Time):
        """
        Updates the failuer status of the circuit breaker

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        pass

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["is_open"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            self.history["is_open"][curr_time] = self.is_open

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def print_status(self):
        """
        Prints the status
        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        pass

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.close()
        if save_flag:
            self.initialize_history()


if __name__ == "__main__":
    pass



"""
This module contains package utils
"""

import numpy as np

from .Time import Time

# Numerical value used as "infinite", choose it based on system values
INF = 1e8


def unique(sequence: list):
    """
    Return list of unique elements while preserving the order

    Parameters
    ----------
    sequence : list
        List of elements

    Returns
    -------
    unq : list
        List of unique elements with preserved order

    """
    seen = set()
    unq = [x for x in sequence if not (x in seen or seen.add(x))]
    return unq


def subtract(list1: list, list2: list):
    """
    Return difference between lists while preserving the order

    Parameters
    ----------
    list1 : list
    list2 : list

    Returns
    -------
    difference : list
        The difference between lists with preserved order

    """

    difference = [x for x in list1 if x not in list2]
    return difference


def intersection(list1: list, list2: list):
    """
    Returns the intersection between two list while preserving the order

    Parameters
    ----------
    list1 : list
        List 1
    list2 : list
        List 2

    Returns
    -------
    intersec : list
        The intersection between two list with preserved order

    """
    intersec = [x for x in list1 if x in list2]

    return intersec


def get_random_instance(seed: int = None):
    """
    Return a numpy random instance with optional seed

    Parameters
    ----------
    seed : int
        Random seed

    Returns
    -------
    random_instance : numpy.random.Generator
        Random instance generator

    """
    random_instance = np.random.default_rng(seed)
    return random_instance


def eq(x: float, y: float, tol: float = 1e-6):
    """
    Checks for equality within a tolerance

    Parameters
    ----------
    x : float
        Left hand side
    y : float
        Right hand side
    tol : float
        The equality tolerance

    Returns
    -------
    equal : bool
        Boolean variable stating whether the left and right hand side
        are equal within the given tolerance

    """
    equal = abs(x - y) < tol
    return equal


def random_choice(random_instance: np.random.Generator, p_true: float):
    """
    Returns random choice based on uniform probability distribution

    Parameters
    ----------
    random_instance : np.random.Generator
        Instance of a random generator
    p_true : float
        The probability of true

    Returns
    -------
    choice : bool
        Random choice based on uniform probability distribution

    """
    if random_instance is None:
        random_instance = get_random_instance()
    choice = random_instance.random() < p_true
    return choice


def interpolate(array: np.ndarray, time_indices: np.ndarray):
    """
    Returns the array interpolated to match the time indices

    arameters
    ----------
    array : np.ndarray
        Array to interpolate
    time_indices : np.ndarray
        Array of required time indices in need of array values

    Returns
    -------
    interpolated_array : np.ndarray
        The array interpolated to match the time indices

    """
    x = np.linspace(0, array.size - 1, time_indices.size)
    interpolated_array = np.interp(x, np.arange(array.size), array)
    return interpolated_array


def convert_yearly_fail_rate(fail_rate_per_year: float, dt: Time):
    """
    Returns random choice based on uniform probability distribution

    Parameters
    ----------
    fail_rate_per_year : float
        The failure rate
    dt : Time
        The current time step

    Returns
    -------
    choice : float
        Random choice
    """
    choice = min(fail_rate_per_year * dt.get_years(), 1)
    return choice


if __name__ == "__main__":
    pass



import numpy as np

from relsad.network.systems import PowerNetwork
from relsad.Time import Time
from relsad.utils import eq


def SAIFI(network: PowerNetwork):
    """
    Returns the current SAIFI (System average interruption failure index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    saifi : float
        The SAIFI value

    """
    interrupted_customers = sum(
        [bus.acc_interruptions * bus.n_customers for bus in network.buses]
    )
    total_customers = sum([bus.n_customers for bus in network.buses])
    if total_customers == 0:
        saifi = 0
    else:
        saifi = interrupted_customers / total_customers
    return saifi


def SAIDI(network: PowerNetwork):
    """
    Returns the current SAIDI (System average interruption duration index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    saidi : float
        The SAIDI value

    """
    sum_outage_time_x_n_customer = sum(
        [
            bus.acc_outage_time.get_hours() * bus.n_customers
            for bus in network.buses
        ]
    )
    total_customers = sum([bus.n_customers for bus in network.buses])
    if total_customers == 0:
        saidi = 0
    else:
        saidi = sum_outage_time_x_n_customer / total_customers
    return saidi


def CAIDI(network: PowerNetwork):
    """
    Returns the current CAIFI (Customer average interruption duration index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    caidi : float
        The CAIDI value

    """
    saifi = SAIFI(network)
    if not eq(saifi, 0):
        caidi = SAIDI(network) / saifi
    else:
        caidi = 0
    return caidi


def ASUI(network: PowerNetwork, current_time: Time):
    """
    Returns the current ASUI (average service unavailability index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element
    current_time : Time
        Current time

    Returns
    ----------
    asui : float
        The ASUI value
    """
    hours = current_time.get_hours()
    asui = SAIDI(network) / hours
    return asui


def ASAI(network: PowerNetwork, current_time: Time):
    """
    Returns the current ASAI (average service avaialbility index)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element
    current_time : Time
        Current time

    Returns
    ----------
    asai : float
        The ASAI value
    """
    asai = 1 - ASUI(network, current_time)
    return asai


def ENS(network):
    """
    Returns the current ENS (Energy not Supplied)

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    ens : float
        The ENS value

    """
    ens = sum([bus.acc_p_energy_shed for bus in network.buses])
    return ens


def EV_Index(network: PowerNetwork):
    """
    Returns the current EV Index, which is the current power demand from
    all the EV parks in the system.

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    ev_index : float
        Index reflecting the current power demand from all the EV parks in
        the system

    """
    ev_index = sum([ev_park.get_ev_index() for ev_park in network.ev_parks])
    return ev_index


def EV_Interruption(network: PowerNetwork):
    """
    Returns the current EV Interruption. Reflects the average number of
    interruptions per EV car for grid support

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    ev_interruption : float
        The average number of interruptions per EV car for grid support

    """
    interruptions_x_num_cars = sum(
        [
            ev_park.acc_exp_interruptions * ev_park.num_cars
            for ev_park in network.ev_parks
        ]
    )
    total_num_cars = sum([ev_park.num_cars for ev_park in network.ev_parks])
    if total_num_cars == 0:
        ev_interruption = 0
    else:
        ev_interruption = interruptions_x_num_cars / total_num_cars
    return ev_interruption


def EV_Duration(network: PowerNetwork):
    """
    Returns the current EV Duration. Reflects the average duration of an EV car
    interruption for grid support

    Parameters
    ----------
    network : PowerNetwork
        A PowerNetwork element

    Returns
    ----------
    ev_duration : float
        The average duration of an EV car interruption for grid support

    """
    sum_interruption_duration = sum(
        [
            ev_park.acc_interruption_duration.get_hours()
            for ev_park in network.ev_parks
        ]
    )
    sum_interruptions = sum(
        [ev_park.acc_num_interruptions for ev_park in network.ev_parks]
    )
    if sum_interruptions == 0:
        ev_duration = 0
    else:
        ev_duration = sum_interruption_duration / sum_interruptions
    return ev_duration



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['relsad',
 'relsad.energy',
 'relsad.examples',
 'relsad.examples.CINELDI',
 'relsad.examples.IEEE16_modified',
 'relsad.examples.IEEE33',
 'relsad.examples.IEEE69',
 'relsad.examples.RBTS2',
 'relsad.examples.RBTS6',
 'relsad.examples.TEST10',
 'relsad.examples.load',
 'relsad.load',
 'relsad.loadflow',
 'relsad.network',
 'relsad.network.components',
 'relsad.network.containers',
 'relsad.network.systems',
 'relsad.reliability',
 'relsad.results',
 'relsad.simulation',
 'relsad.simulation.monte_carlo',
 'relsad.simulation.sequence',
 'relsad.topology',
 'relsad.topology.ICT',
 'relsad.topology.load_flow',
 'relsad.visualization']

package_data = \
{'': ['*'], 'relsad.examples.load': ['data/*']}

install_requires = \
['black>=20.8b1,<21.0',
 'flake8>=3.9.0,<4.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pre-commit>=2.12.0,<3.0.0',
 'scipy>=1.8.0,<2.0.0',
 'sphinx-book-theme>=0.3.2,<0.4.0',
 'sphinxcontrib-bibtex>=2.4.2,<3.0.0']

setup_kwargs = {
    'name': 'relsad',
    'version': '0.0.2',
    'description': 'A package that facilitates reliability investigations in power systems',
    'long_description': '######\nRELSAD\n######\n\n`RELSAD` -- RELiability tool for Smart and Active Distribution networks, is a Python-based\nreliability assessment tool that aims to function as a foundation for reliability\ncalculation of modern distribution systems.\nThe tool uses Monte Carlo simulation and stochastic variation to simulate the\nreliability of a distribution system. The package supports user-selected time\nincrement steps over a user-defined time period. In the tool, active components\nsuch as microgrids, wind power, solar power, batteries, and electrical vehicles\nare implemented. To evaluate smart power systems, ICT components such as\nautomated switches, sensors, and control system for the power grid are also implemented.\nIn addition to component implementation, in order to evaluate the reliability of such\ncomplex systems, the complexity, dependencies within a system, and interdependencies\nbetween systems and components are accounted for.\n\nThe tool can be used in modern distribution network development to evaluate\nthe influence of active components on the network reliability. Relevant use cases\ninclude investigating how:\n\n1. The introduction of microgrids with active production\n   affects the customers in the distribution network and vice versa\n2. Vehicle\\-to\\-grid strategies might mitigate load peaks and\n   improve the distribution network reliability\n3. The reliability of the ICT network impacts the\n   distribution network reliability\n\nExamples using well known test networks are included.\n\n============\nInstallation\n============\n\nSee https://relsad.readthedocs.io/en/latest/installation.html.\n\n========\nFeatures\n========\n\n- Monte Carlo simulation based reliability analysis of active distribution networks\n- Sequential simulation of the network behavoir with user defined loading and failure evolution\n\n============\nDependencies\n============\n\n- Numpy\n- Scipy\n- Matplotlib\n- Pandas\n\n=====\nUsage\n=====\n\nSee https://relsad.readthedocs.io/en/latest/usage.html.\n\n=============\nDocumentation\n=============\n\nThe official documentation is hosted on Read the Docs: https://relsad.readthedocs.io/en/latest/\n\n============\nContributors\n============\n\nWe welcome and recognize all contributions. You can see a list of current contributors in the [contributors tab](https://github.com/stinefm/relsad/graphs/contributors).\n',
    'author': 'Stine Fleischer Myhre',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/stinefm/relsad',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import matplotlib.lines as mlines
import numpy as np

from relsad.Time import Time, TimeUnit

from .Bus import Bus
from .Component import Component
from .Line import Line
from .Switch import Switch


class Disconnector(Component, Switch):

    """
    Common base class for disconnectors

    ...

        Attributes
        ----------
        name : string
            Name of the disconnector
        initial_state : bool
            The initial state of the disconnector (open or colsed)
        is_open : bool
            Tells if the switch is open (True) or closed (False)
        failed : bool
            True if the disconnector is in a failed state, False if not
        fail_rate : float
            The failure rate of the disconnector [no. of fails per year]
        outage_time : Time
            The outage time of the diconnector
        prev_open_time : Time
            The time of the previous time step
        line : Line
            The line the disconnecor is connected to
        base_bus : Bus
            Wich bus the disconnector is closes to (for setting coordinates)
        intelligent_switch : IntelligentSwitch
            Returns the intelligent switch connected to the disconnector if any
        history : dict
            Dictonary attribute that stores the historic variables

        Methods
        ----------
        close()
            Closes the disconnector
        open()
            Opens the disconnector
        fail()
            Sets the disconnector to failed and opens the disconnector
        not_fail()
            Sets the doconnector to not failed and closes the disconnector
        initialize_history()
            Initializes the history variables
        update_history(prev_time, curr_time, save_flag)
            Updates the history variables
        get_history(attribute)
            Returns the history variables of an attribute
        add_random_instance(random_gen)
            Adds global random seed
        reset_status(save_falg)
            Resets and sets the status of the system parameters

    """

    ## Visual attributes
    color = "black"
    edgecolor = "black"
    marker = "o"
    size = 2**2
    handle = mlines.Line2D(
        [],
        [],
        marker=marker,
        markeredgewidth=2,
        markersize=size,
        linestyle="None",
        color=color,
        markeredgecolor=edgecolor,
    )

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        line: Line,
        bus: Bus,
        is_open: bool = False,
        fail_rate: float = 0.014,
        outage_time: Time = Time(1, TimeUnit.HOUR),
    ):

        # Verify input
        if line is None:
            raise Exception("Disconnector must be connected to a Line")
        if line.parent_network is not None:
            raise Exception(
                "Disconnector must be created before the line is connected to a network"
            )
        if bus is None:
            raise Exception("Disconnector must be connected to a Bus")
        if bus not in [line.fbus, line.tbus]:
            raise Exception(
                "Disconnector must be connected to a base bus connected "
                "to the line it is located on"
            )
        if fail_rate < 0:
            raise Exception("The failure rate must be positive")

        self.name = name
        self.initial_state = is_open
        self.is_open = is_open
        self.failed = False
        self.fail_rate = fail_rate
        self.outage_time = outage_time
        self.prev_open_time = Time(0)
        self.line = line

        ## Set coordinate
        self.base_bus = bus
        dx = line.tbus.coordinate[0] - line.fbus.coordinate[0]
        dy = line.tbus.coordinate[1] - line.fbus.coordinate[1]
        if bus == line.tbus:
            dx *= -1
            dy *= -1

        line.disconnectors.append(self)
        self.coordinate = [
            self.base_bus.coordinate[0] + dx / 4,
            self.base_bus.coordinate[1] + dy / 4,
        ]

        ## Communication
        self.intelligent_switch = None

        ## History
        self.history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Disconnector(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, Disconnector)
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def close(self):
        """
        Closes the disconnector

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_open = False
        self.color = "black"
        if not self.line.connected:
            self.line.connect()

    def open(self):
        """
        Opens the disconnector

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_open = True
        self.color = "white"
        if self.line.connected:
            self.line.disconnect()

    def fail(self):
        """
        Sets the disconnecotr to failed and opens the disconnector

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.failed = True
        self.open()

    def not_fail(self):
        """
        Sets the disconnector to not failed and closes the disconnecotr

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.failed = False
        self.close()

    def update_fail_status(self, dt: Time):
        """

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["is_open"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            self.history["is_open"][curr_time] = self.is_open

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def print_status(self):
        """

        Parameters
        ----------

        Returns
        ----------
        None

        """

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.prev_open_time = Time(0)

        self.not_fail()
        if save_flag:
            self.initialize_history()


if __name__ == "__main__":
    pass
